import julia
from julia import Main
from julia import SpecialFunctions

import os
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import time
from tqdm import tqdm
from scipy.io import loadmat
from scipy.stats import norm
from scipy.stats import uniform

#######################################################
# Values that need to be defined ######################
#######################################################

sheetName = 'Fig1D (1D)' #Name of sheet
scanName = 'Fig1D' #Name of scan

startInd = 700 #index to start from
endInd = 1200 #index to end with

plotFlag = True
loops = 1000000

# Priors from Mesa-Fit and information on scan-angle as from previous analysis
angle = 51.77252044 ## Angle of scanning direction in deg
Mst = -24.38345598 ## Mag-thickness product
d = 0.052838539 ## NV-to-sample distance
theta = 59.21594991 ## Out of plane NV orientation
phi = 260.6535864 ## in-plane NV orientations
DWpos = 3.568041957591763236e+01
hdiff = -0.012961466

MstFWHM0 = 1.60968506
dFWHM0 = 0.006727058
thetaFWHM0 = (3.214692023/2.35482)*np.pi/180
phiFWHM0 = (2.540980252/2.35482)*np.pi/180

#################################
# Load Data #####################
#################################

## filepath and excel read
fileDir = os.path.dirname(os.path.abspath(__file__)) ## Determines filedir
excelfilepath=os.path.join(fileDir,'Data.xlsx') ## Path to excelfile
xl = pd.read_excel(excelfilepath,sheetName,usecols="A:B",skiprows=[2]) ## read in excel file
datavals = np.transpose(xl.to_numpy()) ## dataframe extracted from excel file

## assign x,y data
x = datavals[1,startInd:endInd] ## Assign position values
data = datavals[0,1:] ## Assign frequency values
meanval = np.mean(data) ## mean frequency to be subtracted
data = (data-meanval)/(28e-6) ## evaluate stray field from frequency and mean frequency
data = data[startInd:endInd] ## Assign frequency values
Main.x = x


################################
## START OF MHMC ###############
################################


#################################
# Model function implementation #
#################################

#### Julia part to calculate the polygamma function for different arguments
#Main.include("julia.jl")
update_data = Main.eval("""
using SpecialFunctions

function update_data(x,Mst, d,theta,phi,pos,wDW,offset)
    
    z1 = (2 .*d .+ π.*wDW)/(2 .* π .* wDW) .- 1im .*(2 .* (x .- pos))/(2 .* π .* wDW)
    z2 = (2 .*d .+ π.*wDW)/(2 .* π .* wDW) .- 1im .*(2 .* (pos .- x))/(2 .* π .* wDW)
    SpecialFunctions.polygamma.(1,[z1 z2])


end



""")

## Constant prefactor and function def
prf = 1/(2*np.pi**2)

def BlochDW(Mst, d, theta, phi, pos, wDW, offset, x):
    P = update_data(x, Mst, d, theta, phi, pos, wDW, offset)
    P1 = P[:, 0]
    P2 = P[:, 1]
    return offset + prf*(np.cos(theta)*np.imag((1.0/wDW)*Mst*(P1-P2)) - np.cos(phi)*np.real((1.0/wDW)*Mst*(P1+P2))*np.sin(theta))

### STEP 1 INITIALIZE PARAMETERS ##
#### Initialize parameters ########
###################################

nu = len(x)
accepted = 0

###### Check if iterated values do exist already 
filesavepath = os.path.join(fileDir,'MHMC-DW','%sHistappend.txt'%(scanName));
filesavepathFitparas = os.path.join(fileDir,'MHMC-DW','%sFitparameters.txt'%(scanName));

if os.path.isfile(filesavepath):
    loaded = np.loadtxt(filesavepath)
    Mstold=loaded[0]
    dold=loaded[1]
    thetaold=loaded[2]
    phiold=loaded[3]
    posold=loaded[4]
    wDWold=loaded[5]
    offold=loaded[6]
        
    Mst0 = Mstold[-1]
    d0 = dold[-1]
    theta0 = thetaold[-1]
    phi0 = phiold[-1]
    pos0 = posold[-1]
    wDW0 = wDWold[-1]
    off0 = offold[-1]
else:
    ## If Fit does not exist yet. Define the starting values based on the given parameters
    if hdiff >0:
        d0 = d+hdiff
    else:
        d0 = d
    ### Take the absolute value of the magnetization
    Mst0 = Mst
    theta0 = theta*np.pi/180
    ### Adjust the inplane NV angle according to scan direction
    phi0 = phi*np.pi/180 - angle
    pos0 = DWpos
    wDW0 = 0.05
    off0 = 1

######################################################
# Define the Priors and propability density function #
######################################################

###### Predefined centers and width of the distributions.
## First 4 are normal distributions. remaining 3 just bounds
## dist 1 is the center values. dist 2 is the uncertainty // dist1 is Lower Bound and dist2 is Upper Bound

posLB = -200
posUB = 200
wDWLB = 0
wDWUB = 0.5
offLB = -20
offUB = 20

## define prior centers. 
Mstprc = Mst0 
dprc = d0
thetaprc = theta0
phiprc = phi0

dist1 = [Mstprc,dprc,thetaprc,phiprc,posLB,wDWLB,offLB]
dist2 = [MstFWHM0, dFWHM0,thetaFWHM0, phiFWHM0, posUB,wDWUB, offUB]


## Def Propability density function (pdf) from the Priors and Likelyhood #################
def pdf(param,x,data,dist1,dist2):
    x1 = BlochDW(*param,x)

    ### Calculate the log of Priors and likelyhood. Uniform for pos,wDW and offset, else normally distributed
    MstPr, dPr, thetaPr, phiPr = norm.logpdf([(param[0]-dist1[0])/dist2[0]*2.35482,(param[1]-dist1[1])/dist2[1]*2.35482,(param[2]-dist1[2])/dist2[2]*2.35482,(param[3]-dist1[3])/dist2[3]*2.35482])
    posPr,wDWPr,offPr = uniform.logpdf([param[4],param[5],param[6]],[dist1[4],dist1[5],dist1[6]],[dist2[4],dist2[5],dist2[6]])

    ## is the prior returns a very large negative value return a smaller value instead.
    if posPr < -1e10: posPr = -1e10
    if wDWPr < -1e10: wDWPr = -1e10
    if offPr < -1e10: offPr = -1e10

    ### RETURN LOG of LIKELYHOOD AND PRIORS. RATIO OF LOGS AFTERWARDS IS THE DISTANCE BETWEEN THE LOGS.
    y = -((nu+1)/2)*np.log(np.sum((x1-data)**2)) + MstPr + dPr + thetaPr + phiPr + posPr + wDWPr + offPr
    return y



#######################################################
#######################################################
## Metropolis Hastings Iteration for n=loops LOOPS ####
#######################################################
#######################################################


# Create the arrays needed for the iterations and define first element as initial vals# 

theA = np.log(np.random.random(loops)) # Log of Random number for acceptance criteria
Mst = np.zeros(loops+1) # Parameter 1: Mag-thickness product
Mst[0] = Mst0 
d = np.zeros(loops+1) # Parameter 2: NV to sample distance
d[0] = d0
theta = np.zeros(loops+1) # Parameter 3: Theta-angle of NV sensor
theta[0] = theta0
phi = np.zeros(loops+1) # Parameter 4: Phi-angle of NV sensor
phi[0] = phi0
pos = np.zeros(loops+1) # Paramterer 5: Position of DW
pos[0]=pos0
wDW = np.zeros(loops+1) # Parameter 6: Width parameter of DW
wDW[0] = wDW0
off = np.zeros(loops+1) # Parameter 7: offset on data
off[0] = off0

# Define the Step Sizes #

MstStep = abs(0.001*Mstprc)
dStep = abs(0.001*dprc)
thetaStep = abs(0.001*thetaprc)
phiStep = abs(0.001*phiprc)
posStep = 0.001
wDWStep = 0.001
offStep = 0.001

# START LOOP #

for i in tqdm(range(0,loops)):
    Mst_p, d_p, theta_p, phi_p, pos_p, wDW_p, off_p = np.random.normal([Mst[i], d[i], theta[i],phi[i],pos[i],wDW[i],off[i]],[MstStep,dStep,thetaStep,phiStep,posStep,wDWStep, offStep])
    post = [Mst_p,d_p,theta_p,phi_p,pos_p,wDW_p,off_p]
    init = [Mst[i],d[i],theta[i],phi[i],pos[i],wDW[i],off[i]]
    diff = pdf(post,x,data,dist1,dist2)-pdf(init,x,data,dist1,dist2)
    if theA[i]<=diff: ### Accept the porposed candidate parameterset
            Mst[i+1] = Mst_p
            d[i+1] = d_p
            theta[i+1] = theta_p
            phi[i+1] = phi_p
            pos[i+1] = pos_p
            wDW[i+1] = wDW_p
            off[i+1] = off_p
            accepted += 1 ## increase number of accepted points by 1
    else: ### If new proposed candidate paramterset is rejected take previous values
            Mst[i+1] = Mst[i]
            d[i+1] = d[i]
            theta[i+1] = theta[i]
            phi[i+1] = phi[i]
            pos[i+1] = pos[i]
            wDW[i+1] = wDW[i]
            off[i+1] = off[i]

# Print acceptance rate #
print('Acceptance rate was ', (accepted/loops)*100, '%')



#################################
#################################
### Plot and Give the output ####
#################################
#################################


# Determine median for Fit parameters. 
# This is another estimate and often identical to a Gaussian Fit within uncertainty.
# For final analysis it is advised to fit the distributions by a Gaussian
if os.path.isfile(filesavepath):
    Mst = np.append(Mstold,Mst)
    d = np.append(dold,d)
    theta = np.append(thetaold,theta)
    phi = np.append(phiold,phi)
    pos = np.append(posold,pos)
    wDW = np.append(wDWold,wDW)
    off = np.append(offold,off)
    print(len(Mst))
else:
    print(len(Mst))

parsf = [np.median(Mst), np.median(d), np.median(theta), np.median(phi), np.median(pos), np.median(wDW), np.median(off)]
print(parsf)

np.savetxt(filepathsave,[Mst,d,theta,phi,pos,wDW,off])
np.savetxt(filesavepathFitparas, parsf)

### Show plots and save of the parameterdevelopment #####
paramdevelFlag = True

if paramdevelFlag:
    fig = plt.figure(figsize=(11,8))
    plt.plot(Mst/Mst0)
    plt.plot(d/d0)
    plt.plot(theta/theta0)
    plt.plot(phi/phi0)
    plt.plot(pos/pos0)
    plt.plot(wDW/wDW0)
    plt.plot(off/off0)
    plt.legend(['Mst','d','theta','phi','pos','wDW','off'], loc=2)
    plt.show()

### Show plot and save of The individual histograms
binno = 30
fig = plt.figure(figsize=(11,8))

plt.subplot(2,4,1)
plt.hist(Mst[int(loops/2):],'auto')
plt.xlabel('Magnetization')

plt.subplot(2,4,2)
plt.hist(d[int(loops/2):], 'auto')
plt.xlabel('NV dist')

plt.subplot(2,4,3)
plt.hist(theta[int(loops/2):]*180/np.pi, 'auto')
plt.xlabel('Theta')

plt.subplot(2,4,4)
plt.hist(phi[int(loops/2):]*180/np.pi, 'auto')
plt.xlabel('Phi')

plt.subplot(2,4,5)
plt.hist(pos[int(loops/2):], 'auto')
plt.xlabel('DW Position')

plt.subplot(2,4,6)
plt.hist(wDW[int(loops/2):], 'auto')
plt.xlabel('DW Width')

plt.subplot(2,4,7)
plt.hist(off[int(loops/2):], 'auto')
plt.xlabel('Offset')

plt.subplot(2,4,8)
plt.plot(x,data,'ro')
plt.xlabel('Pos (um)')
plt.ylabel('Field (uT)')
plt.title('Scatter Plot of Data')
plt.plot(x,BlochDW(*parsf,x))

plt.show()


fig = plt.figure(figsize=(11,8))
plt.plot(x,data,'ro')
plt.plot(x,BlochDW(parsf[0],parsf[1],parsf[2],parsf[3],parsf[4],parsf[5],parsf[6],x))
plt.legend(['data','Bloch'], loc=2)
plt.show()
